(****************** Matematica - corso di LPL: Laboratorio 2020 *****************)
(* ******************************************************************************)
(* Definizione del Linguaggio Small2020: Nucleo di C modificato ed esteso       *)
(* Realizzazione di un Esecutore di Small2020 utilizzando OCaml                 *)
(*                + Laboratorio del 03/04/20                                    *)
(*                + Laboratorio del 07/04/20                                    *)
(*                + Laboratorio del 08/04/20                                    *)
(*                + Laboratorio del 09/04/20                                    *)
(*                + Laboratorio del 19/04/20                                    *)
(*                + Laboratorio del 24/04/20                                    *)
(*                + Laboratorio del 05/05/20                                    *)
(*                + Laboratorio del 08/05/20                                    *)
(*                + Da copiare, in Small5, prima di modificare   			    *)
(********************************************************************************)

(* =============================================================================*)
(* Parametri di Programma e Globali *)

let storeSize = 1000;;   (* store size in int/bool words *)


(* ============================================================================ *)
(* The Token, i.e. Lexical Categories *)

type ide = string;;
type num = int;;
type bool = True | False;;



(* ============================================================================ *)
(* The Syntactic Categories *)
(* Vedi Laboratorio2 (slides), laboratorio1: Ocaml del 27/03/20 *)

type tye = 
	Int
	| Bool
	| Void
	| Arr of tye * num
	| Mut of tye
	| Abs of tye * tyeSeq
	| Terr
	and

tyeSeq = tye list
	and	
	
dcl = 
	Const of tye * ide * exp   (* corretta vedi AST *)
	| Var of tye * ide * exp
	| VarN of tye * ide
	| Array of tye * ide * num
	and 
	
exp = 
	Val of ide
	| N of num
	| B of bool
	| GetArrow of ide * exp
	| Neg of exp
	| Plus of exp * exp
	| Minus of exp * exp
	| Mul of exp * exp
	| Div of exp * exp
	| Eq of exp * exp
	| LT of exp * exp
	| GT of exp * exp
	| Not of exp
	| Or of exp * exp
	| And of exp * exp
	| Upd of exp * exp
	| EE
	and 

cmd = 
	SeqC of cmd * cmd
	| IfE of exp * cmd * cmd
	| IfT of exp * cmd
	| For of stm * exp * stm * cmd
	| Cmd of exp
	| EC
	and 
	
stm = 
	StmD of dcl
	| StmC of cmd
	| Seq of stm list
	;;			

type prog = Prog of ide * stm;;

(* Operations for the AST presentation in the given Concrete Syntax *)
exception Error of string;;

open Printf;; 

let toStringI (i:ide) = sprintf "%s" i;;

let toStringN (n:num) = sprintf "%d" n;;

let toStringB = function
		| True -> "true"
		| False -> "false"
		;;

let rec indent tab = 
		if tab < 0 then ""
		else if tab = 0 then ""
			 else " " ^ (indent (tab-1))
		;;

let rec toStringTye = (function 
		| Int -> "int"
		| Bool -> "bool"
	    | Void -> "void"
	    | Arr(ty,n) -> (toStringTye ty)^ "[" ^(toStringN n)^ "]"
	    | _ -> let msg = "toStringTye: Hide types in AST" in
	    	   raise (Error msg))	
       and

toStringTseq x = (
		let msg = "toStringTye: Hide types in AST" in
	    raise (Error msg)) 
		and 

toStringDcl = (function
		| Const(tye,ide,exp) -> ("final " ^ (toStringTye tye) ^ " " 
								 ^ (toStringI ide) ^ " = " ^ (toStringExp exp))
								 (* corretto arg. exp/toStringExp - vedi AST *)
		| Var(tye,ide,exp) -> ("var " ^ (toStringTye tye) ^ " " 
							   ^ (toStringI ide) ^ " = " ^ (toStringExp exp))
		| VarN(tye,ide) -> ("var " ^ (toStringTye tye) ^ " " ^ (toStringI ide))
		| Array(tye,ide,num) -> ((toStringTye tye) ^ " array " ^ " " 
								 ^ (toStringI ide) ^ "[" ^ (toStringN num) ^ "]"))
		and

toStringExp = (function
		| Val ide -> toStringI ide
	    | N num -> toStringN num
	    | B bool -> toStringB bool
		| GetArrow(ide,exp) -> (toStringI ide) ^ "[" ^ (toStringExp exp) ^ "]"
		| Neg exp -> "-" ^ (toStringExp exp) 
		| Plus(exp1,exp2) -> "("^(toStringExp exp1) ^ " + " ^ (toStringExp exp2)^")"
		| Minus(exp1,exp2) -> "("^(toStringExp exp1) ^ " - " ^ (toStringExp exp2)^")"
		| Mul(exp1,exp2) -> "("^(toStringExp exp1) ^ " * " ^ (toStringExp exp2)^")"
		| Div(exp1,exp2) -> "("^(toStringExp exp1) ^ " / " ^ (toStringExp exp2)^")"
		| Eq(exp1,exp2) -> "("^(toStringExp exp1) ^ " == " ^ (toStringExp exp2)^")"
		| LT(exp1,exp2) -> "("^(toStringExp exp1) ^ " < " ^ (toStringExp exp2)^")"
		| GT(exp1,exp2) -> "("^(toStringExp exp1) ^ " > " ^ (toStringExp exp2)^")"
		| Not exp -> "not " ^ (toStringExp exp)
		| Or(exp1,exp2) -> "("^(toStringExp exp1) ^ " or " ^ (toStringExp exp2)^")"
		| And(exp1,exp2) -> "("^(toStringExp exp1) ^ " and " ^ (toStringExp exp2)^")"
		| Upd(exp1,exp2) -> "("^(toStringExp exp1) ^ " = " ^ (toStringExp exp2)^")"
		| EE -> " ")
		and
		
isAtomic = (function  
		| Val _ -> true
	    | N _ -> true
	    | B _ -> true
		| GetArrow _ -> true
		| Neg _ -> true
		| Not _ -> true
		| _ -> false)
		and

toStringCmd tab = (function
		| SeqC(cmd1,cmd2) -> ((indent tab) ^ "{\n" ^ (flattenC (tab+1) cmd1) 
							  ^ (flattenC (tab+1) cmd2) ^ (indent tab) ^ "}")
		| IfE(exp,cmd1,cmd2) -> (let expString = 
									 (if (isAtomic exp) 
									  then "(" ^ (toStringExp exp) ^ ")" 
									  else toStringExp exp) in
								 (indent tab) ^ "if " ^ expString 
							      ^ (toStringCmd 1 cmd1) ^ ";\n" 
							      ^ (indent tab) ^ "else " ^ (toStringCmd 0 cmd2))
		| IfT(exp,cmd) -> (let expString = 
							   (if (isAtomic exp) 
								then "(" ^ (toStringExp exp) ^ ") " 
								else toStringExp exp) in
						   (indent tab) ^ "if " ^ expString ^ (toStringCmd 1 cmd))
		| For(stm1,exp,stm2,cmd) -> ((indent tab) ^ "for (" ^(toStringStm 0 stm1) ^ ";" 
							     ^ (toStringExp exp) ^ ";" 
							     ^ (toStringStm 0 stm2) ^ ")\n" 
							     ^ (toStringCmd (tab+10) cmd))
		| Cmd exp -> (let s = toStringExp exp in
				      let expString = ( 
					  if isAtomic exp then s
				      else String.sub s 1 ((String.length s)-2)) in
		              (indent tab) ^ expString)
		| EC -> " " )
		and
		
flattenC tab = (function
		| SeqC(cmd1,cmd2) -> (flattenC tab cmd1) ^ (flattenC tab cmd2)
		| EC -> "" 
		| x -> (toStringCmd tab x) ^ ";\n")
		and
		
toStringStm tab = (function
		StmD dcl -> (indent tab) ^ (toStringDcl dcl)
		| StmC cmd -> toStringCmd tab cmd
		| Seq list -> let g = fun s -> (toStringStm tab s)^";\n" in
				  	  let ll = List.map g list in
				      List.fold_left (^) "" ll)
	    and
	
toStringProg (Prog(ide,stm)) = 
		let body = toStringStm 3 stm in
		"Program " ^ (toStringI ide) ^ "{\n" ^ body ^ "}"
				
let printDcl dcl = printf "\n%s\n\n" (toStringDcl dcl);;
let printExp exp = printf "\n%s\n\n" (toStringExp exp);;
let printCmd tab cmd = printf "\n%s\n\n" (toStringCmd tab cmd);;
let printStm tab stm = printf "\n%s\n\n" (toStringStm tab stm);;
let printProg prog = printf "\n%s\n\n" (toStringProg prog);;
		
		
(* ============================================================================ *)
(* The Semantic Structure: Store e Environment *)	
(* Store: Definition *)
(********** Store: Structure *)
type loc = Loc of int;;
type mval = Mint of int 
	| Mbool of bool
	| Undef
	;;
type store = Store of int * (loc -> mval);;
(* Funzione di Astrazione ed Invariante di Rappresentazione per store: 
1) Presentazione (o forma dei valori): [l1<-Mv1,...,lk<-Mvk] per k>=0
2) AF(c) = [] if c=Store(0,g) 
   AF(c) = [l1<-Mv1,...,lk<-Mvk] 
   		   if c=Store(k,g) and (perOgni i\in[1..k], li=Loc(i-1) and (g li)=Mvi).
3) I(Store(n,g)) = 0≤n<=storeSize and 
		           (perOgni i\in[n+1..storeSize], li=Loc(i-1) and (g li) = Undef
*) 
(********** Store: Operations and Exceptions *)
exception IllegalAddress of string * loc;;
exception UndefinedLoc of string * loc;;
exception StoreSizeExceeded of string * int * int;;

let emptyStore () =  
		Store (0 , fun ((Loc n)as x) -> 
					if (n >= 0) && (n < storeSize) 
					then Undef
					else raise(IllegalAddress("emptyStore",x)))
		;; (* uso di patterns come parametri formali e di "as" per nominarli *)
		
let allocate (Store(d,g)) (k:int) = 
		if (d+k<=storeSize)
		then ((Loc d),Store(d+k,g))
		else raise (StoreSizeExceeded ("allocate",d,k))
		;; (* uso di pattern come formali e di formali con tipo *)

let checkLoc = fun (Loc x) d -> x >= 0 && x < d ;; 	(* operazione locale *)	

let upd (Store(d,g)) (loc:loc) (mv:mval) =  (* DA FAR PROGRAMMARE *)
		if checkLoc loc d
		then Store(d, fun l -> if l = loc then mv else g l)
		else raise (IllegalAddress("upd",loc))
		;; (* uso di pattern come formali e di formali con tipo *)

let getStore (Store(d,g)) (loc:loc) =  (* DA FAR PROGRAMMARE *)
		if not(checkLoc loc d) 
		then raise (IllegalAddress("getStore",loc))
		else match g loc with
				Undef -> raise (UndefinedLoc ("getStore",loc))
				| ret -> ret
				;;(* uso di pattern come formali e di formali con tipo *)

(********** Store: Operation for Store Presentation *)
let toStringMval = function
		Mint n -> sprintf "%d" n 
		| Mbool b -> sprintf "%s" (toStringB b)
		| _ -> "Undef";;

let toStringLoc (Loc n) = sprintf "L%d" n;;

let toStringStore (Store(d,g)) = (* una presentazione dello store *)
		if d = 0 then "[]"
		else let pair n h = 
						(let loc = Loc n in 
						 let v = toStringMval(h loc) and l = toStringLoc loc in
						 l^"<-"^v) in
		     let rec range n m h = (* requires n<m *)
						(if n = m-1 
						then pair n h
						else (pair n h)^","^(range (n+1) m h)) in	
			"["^(range 0 d g)^"]"							
		;;
let storeDump store = 			 (* dump dello store *)
printf "\n%s\n\n" (toStringStore store);;
						
(* Environment: Definition *)
(********** Env: Structure *)
type dval = 
	DConstN of tye * num
	| DConstB of tye * bool
	| DVar of tye * loc
	| DArray of tye * loc
	| Unbound
	;;
type env = Env of ide list * (ide -> dval);;
(* Funzione di Astrazione ed Invariante di Rappresentazione per env.
1) Presentazione (o forma dei valori): [id1/D1,...,idk/Dk] per k>=0
2) AF(c) = [] if c = Env(l,g) and (perOgni id, g(id) = Unbound)
   AF(c) = [id1/D1,...,idk/Dk] 
   		 if c = Env(l,g) and (g(idi) = Di per i\in[1..k], 
   		                      g(x) = Unbound altrimenti).
3) I(c) = (c = Env(l,g) => (List.mem x l) iff (g(x) != Unbound)
*)
(*********** Env: Operations and Exceptions *)
exception UnboundIde of string * ide;;
exception WrongDval of string

open List;;
open String;;

let isSimple = function
	Int -> true
	| Bool -> true
	| _ -> false
	;; 

let isDval = function
	DConstN (Int,(n:num)) -> true
    | DConstB (Bool,(b:bool)) -> true
	| DVar(Mut t,loc) -> isSimple t
			(* corretto primo arg.: t con Mut t *) 
	| DArray(Mut(Arr(Mut t,n)),loc) -> isSimple t
			(* corretto primo arg.: t con Mut t *) 
	| Unbound -> true
	| _ -> false
	;;
		
let emptyEnv () = 
		Env ([] , fun ide -> Unbound)
		;;
		
let bind (Env(l,r)) ide dval =
		if isDval dval 
		then let g = fun i -> if i = ide then dval else r i in
		     Env(ide::l , g)
	    else raise(WrongDval("toStringDval"))
		;;
			
let defined (Env(l,r)) ide =
		mem ide l
		;;
		
let getEnv (Env(l,r)) ide =
		if mem ide l then r ide
		else raise (UnboundIde ("getEnv",ide))
		;;

(* ******** Env: Operation for env presentation *)

let toStringDval = function
		DConstN (Int,(n:num)) -> sprintf "(int,%d)" n
		| DConstB (Bool,(b:bool)) -> sprintf "(bool,%s)" (toStringB b)
		| DVar(Mut t,loc) 
				when isSimple t 
				-> sprintf "(%s,%s)" (toStringTye t) (toStringLoc loc)
				(* corretto primo arg.: t con Mut t *) 
		| DArray(Mut(Arr(Mut t,n)),loc) 
				when isSimple t
				-> sprintf "(%s,%s)" (toStringTye (Arr(t,n))) 
									 (toStringLoc loc)
				(* corretto primo arg.: (Arr..) con Mut (Arr..) *)		  
		| Unbound -> "Unbound"
		| _ -> raise(WrongDval("toStringDval"))
		;;

let toStringEnv (Env (ideList, g)) = (* una presentazione di Env *)
		let pair h i = (let ide = toStringI i and 
		                    den = toStringDval (h i) in ide^"/"^den) in
        let pList = List.map (pair g) ideList in
        let gTemp = (fun x a -> let check = ((get a 0) = ']') in
        			            if check then x^a else (x^"; "^a)) in
        "[" ^ (fold_right gTemp pList "]")
		;;

let envDump env = 			(* dump di env *)
		printf "\n%s\n\n" (toStringEnv env)
		;;

(*********** State: A pair (env,store) -- Operation for state presentation *)
let showState (env,store) =           (* dump dello stato *)
		printf "\nStack:\n%s\nStore:\n%s\n\n" (toStringEnv env) 
											  (toStringStore store) 
		;;

let sigma0 = (emptyEnv(),emptyStore())
		;;



(* ============================================================================ *)
(* The Semantic Function: dclSem, expSem, cmdSem, progSem *)
exception TypeErrorI of string * ide;;
exception SystemError of string * ide;;
exception SystemErrorM of string;;
exception TypeErrorE of string;;
exception SystemErrorE of string;;
exception StaticErrorE of string * ide;;

type eval = 
	Ival of num
	| Bval of bool
	| E
	;;
	
let eTOm = function
	Ival n -> Mint n
	| Bval b -> Mbool b
	| _ -> raise(SystemErrorE("eTOm"))
	;;
	
let mTOe = function
	Mint n -> Ival n
	| Mbool b -> Bval b
	| _ -> raise(SystemErrorM("mTOe"))
	;;
	
let ysame t1 t2 = t1 = t2
    ;;
    
(* ============================================================================ *)
(* The Semantic Function: dclSem *)
    
let rec dclSem dcl (rho,(Store(d,g)as mu)) = 
	match dcl with
	  |Const(ty,ide,exp) -> 
		    (match expSem exp (rho,mu) with
			   |(te,v,(rho1,mu1)) 
			   	  	  when (isSimple te) && (ysame ty te)
			   	  	  -> (let den = (
			   	  	          function 
			   	  	  			|(Int,Ival n )-> DConstN(Int,n)
			   	  	  			|(_,Bval b) -> DConstB(Bool,b)
			   	  	  			|_ -> raise(SystemError("dclSem:Const",ide))) in
			   	  	      let rho2 = bind rho1 ide (den(te,v)) in
			   	  	      let st2 = (rho2,mu1) in
			   	  	      (Void,st2))
			   	|(te,v,(rho1,mu1))
			   		   when (isSimple te) 
			   		   -> raise(TypeErrorI("E2: dclSem",ide))
			   	|_ -> raise(TypeErrorI("E1: dclSem",ide)))
		| Var(ty,ide,exp) -> 
		    (match expSem exp (rho,mu) with
			   |(te,v,(rhoe,mue)) 
			   	  	  when (isSimple te) && (ysame ty te)
			   	  	  -> (let (loca,mua) = allocate mue 1 in
			   	  	      let den = DVar((Mut ty),loca) in
			   	  	      let rhoF = bind rhoe ide den in
			   	  	      let muF = upd mua loca (eTOm v) in
			   	  	      (Void,(rhoF,muF)))
			   	|(te,v,(rhoe,mue))
			   		   when (isSimple te) 
			   		   -> raise(TypeErrorI("E2: dclSem",ide))
			   	|_ -> raise(TypeErrorI("E1: dclSem",ide)))
		| VarN(ty,ide) 
				when (isSimple ty)
				-> (let(loca,muF) = allocate mu 1 in
					let den = DVar((Mut ty),loca) in
					let rhoF = bind rho ide den in
					(Void,(rhoF,muF)))
		| VarN(ty,ide)
			   	-> raise(TypeErrorI("E3: dclSem",ide))		        
		| Array(t,ide,num) 
				when (isSimple t) && (num > 0)
	    	   	-> (let (loca,muF) = allocate mu num in
	    	   	   let den = DArray(Mut(Arr(Mut t,num)),loca) in
	    		   let rhoF = bind rho ide den in
	    		   (Void,(rhoF,muF)))
		| Array(t,ide,num) 
				when (num > 0)
			   	-> raise(TypeErrorI("E6: dclSem",ide))
		| Array(t,ide,num) 
			   	-> raise(TypeErrorI("E7: dclSem",ide))
		and 


(* ============================================================================ *)
(* The Semantic Function: expSem *)
expSem exp (rho,(Store(d,g)as mu)) = 
	match exp with
	  | N n -> 
	 	 (Int,Ival n,(rho,mu))
	  | B b -> 
	 	 (Bool,Bval b,(rho,mu))
	  | EE -> 
	 	 (Void,E,(rho,mu))
	  |_ -> let msg = (toStringExp exp) ^ " can't be used" in 
	        raise(SystemErrorE("expSem: " ^ msg))
	  and

dexpSem dexp (rho,(Store(d,g)as mu)) =
	match dexp with
	  |Val ide -> (
	  	try(let den = getEnv rho ide in
	  		match den with
	  		 |DVar(Mut t,loct) when (isSimple t)
	  		    -> (t,loct,(rho,mu))
	  		 |DVar(Mut t,loct)
	  		    -> raise(TypeErrorE("E8: dexpSem - " ^ (toStringExp dexp)))
	  		 |_ -> raise(TypeErrorE("E9: dexpSem - " ^ (toStringExp dexp)))
	  	) with 
	  		| (UnboundIde _)
	  			 -> (let msg = "dexpSem: Unbound Identifier " in
	  			     raise (StaticErrorE(msg)))
	  )
	  |_ -> let msg = (toStringExp dexp) ^ " can't be used" in 
	        raise(SystemErrorE("expSem: " ^ msg))
	  ;;
		
(* ============================================================================ *)
(* The Semantic Function: cmdSem *)


(* ============================================================================ *)
(* The Semantic Function: progSem *)


(* ============================================================================ *)
(* Tests: Abstract Syntax *)
(* tests in Laboratorio3 *)
(* Program P0 *)

(* ============================================================================ *)
(* Tests: Semantic Structures *)
(* tests in Laboratorio4 *)
(******** Tests on Store *)
(******** Tests on Env *)

(* ============================================================================ *)
(* Tests: Semantic Function dclSem *)
(* Tests in Laboratorio5 *)

(* ============================================================================ *)
(* Tests: Semantic Function expSem *)
(*1 - Verificare che expSem calcola correttamente nelle dichiarazioni d, d1 sotto:
d = Const(Int,"luigi",N 7) e d1 = Var(Bool,"piero",B False) 
let d = Const(Int,"luigi",N 7) and d1 = Var(Bool,"piero",B False) in
let (t,st) = dclSem d sigma0 in
let (t1,st1) = dclSem d1 st in
showState st1;;
*)
(* 2 - Proporre un Test (nota, al momento possiamo usare solo dichiarazioni) 
let d = Const(Int,"luigi",N 7) and d1 = Var(Bool,"piero",B False) 
        and d2 = Var(Int, "x", Val "luigi") in
let (t,st) = dclSem d sigma0 in
let (t1,st1) = dclSem d1 st in
let (t2,st2) = dclSem d2 st1 in
showState st2;;  *)
(*
Sostituire, d2 con 
d2 = Var(Bool, "x", Val "luigi") ed ancora con 
d2 = Var(Bool, "x", EE)
*)
(* 3 - Proporre un Test: *)

(* 4 - Proporre un Test: *)



(* Tests: Programs and Computations *)
(* Tests: State Dumping *)



